@@ -18,6 +18,12 @@ use pyo3::{
     types::{PyBytes, PyTuple},
 };
 use rand::thread_rng;
+use rayon::iter::IndexedParallelIterator;
+use rayon::iter::ParallelIterator;
+use rayon::{
+    prelude::{IntoParallelIterator, IntoParallelRefIterator},
+    slice::ParallelSlice,
+};
 use wide_ciphertexts::WideCiphertext;
 
 // TODO: Consider returning Vecs instead of bytes
@@ -52,16 +58,16 @@ fn build_okvs(py: Python, public_key: PyObject, strings_in_each_row: Vec<Vec<Str
     let public_key_table = EdwardsBasepointTable::create(&read_public_key(py, public_key));
 
     let data_rows: Vec<(DataRow, Ciphertext)> = strings_in_each_row
-        .into_iter()
+        .into_par_iter()
         .map(|row| (DataRow(row), Ciphertext::zero(&public_key_table)))
         .collect();
 
+    // TODO: Parallelize Paxos
     let okvs: Paxos<Ciphertext> = Paxos::encode(&data_rows, 40);
 
     PyBytes::new(py, &okvs.to_bytes()).into()
 }
 
-// TODO: Parallelize
 #[pyfunction]
 fn initiate_queries(
     py: Python,
@@ -76,7 +82,7 @@ fn initiate_queries(
         .expect("The sender_okvs object should be of type `bytes`.");
     let sender_okvs_instance: Paxos<Ciphertext> = Paxos::from_bytes(sender_okvs_bytes.as_bytes());
     let sender_ciphertexts = sender_data
-        .into_iter()
+        .into_par_iter()
         .map(|row| sender_okvs_instance.decode(&DataRow(row)));
 
     let receiver_okvs_bytes: &PyBytes = receiver_okvs
@@ -85,7 +91,7 @@ fn initiate_queries(
     let receiver_okvs_instance: Paxos<Ciphertext> =
         Paxos::from_bytes(receiver_okvs_bytes.as_bytes());
     let receiver_ciphertexts = receiver_data
-        .into_iter()
+        .into_par_iter()
         .map(|row| receiver_okvs_instance.decode(&DataRow(row)));
 
     let public_key = read_public_key(py, swift_pk);
@@ -104,7 +110,7 @@ fn randomize(py: Python, ciphertexts: PyObject) -> PyObject {
         .expect("The ciphertexts object should be of type `bytes`.");
     let randomized_ciphertexts: Vec<u8> = ciphertexts_bytes
         .as_bytes()
-        .chunks(128)
+        .par_chunks(128)
         .flat_map(|chunk| WideCiphertext::from_bytes(chunk).randomize().to_bytes())
         .collect();
 
@@ -120,21 +126,17 @@ fn combine(py: Python, sender_ciphertexts: PyObject, receiver_ciphertexts: PyObj
         .cast_as(py)
         .expect("The receiver_ciphertexts object should be of type `bytes`.");
 
-    let mut alpha_bytes = vec![];
-    let mut beta_bytes = vec![];
-    let mut gamma_bytes = vec![];
-    let mut delta_bytes = vec![];
-    for (x, y) in sender_ciphertexts_bytes
+    let wide_ciphertexts: Vec<WideCiphertext> = sender_ciphertexts_bytes
         .as_bytes()
-        .chunks(128)
-        .zip(receiver_ciphertexts_bytes.as_bytes().chunks(128))
-    {
-        let ciphertext = WideCiphertext::from_bytes(x).sum(WideCiphertext::from_bytes(y));
-        alpha_bytes.extend(ciphertext.a);
-        beta_bytes.extend(ciphertext.b);
-        gamma_bytes.extend(ciphertext.c);
-        delta_bytes.extend(ciphertext.d);
-    }
+        .par_chunks(128)
+        .zip(receiver_ciphertexts_bytes.as_bytes().par_chunks(128))
+        .map(|(x, y)| WideCiphertext::from_bytes(x).sum(WideCiphertext::from_bytes(y)))
+        .collect();
+
+    let alpha_bytes: Vec<u8> = wide_ciphertexts.par_iter().flat_map(|c| c.a).collect();
+    let beta_bytes: Vec<u8> = wide_ciphertexts.par_iter().flat_map(|c| c.b).collect();
+    let gamma_bytes: Vec<u8> = wide_ciphertexts.par_iter().flat_map(|c| c.c).collect();
+    let delta_bytes: Vec<u8> = wide_ciphertexts.par_iter().flat_map(|c| c.d).collect();
 
     let sender_points: PyObject = PyBytes::new(py, &alpha_bytes).into();
     let receiver_points: PyObject = PyBytes::new(py, &beta_bytes).into();
@@ -161,7 +163,7 @@ fn decrypt(py: Python, points: PyObject, bank_secret_key: PyObject) -> PyObject
         .expect("The points object should be of type `bytes`.");
     let decrypted_points: Vec<u8> = points_bytes
         .as_bytes()
-        .chunks(32)
+        .par_chunks(32)
         .flat_map(|chunk| {
             (secret_key * CompressedEdwardsY::from_slice(chunk).decompress().unwrap())
                 .compress()
